//  *****************************************************************************************************************
//  *                                                                                                               *
//  *                                         SpikenzieLabs.com                                                     *
//  *                                                                                                               *
//  *                                           Calculator Kit                                                      *
//  *                                                                                                               *
//  *****************************************************************************************************************
//
// BY: MARK DEMERS - SpikenzieLabs
// Dec 2012
// VERSION 1.1
//
// Brief:
// Sketch used in the SpikenzieLabs Calculator Kit, more info and build instructions at http://www.spikenzielabs.com/calc
// 
//
// LEGAL:
// This code is provided as is. No guaranties or warranties are given in any form. It is your responsibilty to 
// determine this codes suitability for your application.

// The SpikenzieLabs Calculator Kit software and hardware is licensed under Attribution-NonCommercial.  If you would like to 
// make a commercial product based on our works in whole or in part, please contact liscencing@spikenzielabs.com to discuss 
// possible licensing arrangements.
//
// NOTICE:
// SpikenzieLabs does not permit use of this device or code to calculate spacecraft trajectories, other mission critcal
// calculations or life support uses. This is a novelty calculator and even thou every attempt has been made it
// calculate accurately it's processor is limited to a maximum of 6 or 7 digits of "reasonable" accuracy.
// 
// Changes:
// v0  - Float numbers test - for future calculator Kit
// v1c - Started to add decimal point implementation
// v1E - Changed INT to long formate (needed larger numbers)
// v1f - Tring to reduce variable sizes (have some strange results - may be too much SRAM memory usage)
//     - ie: Changing int variable to byte
// v1f2 - Removed Long interger math
// v1f3 - Removed most local variable (reusing globals in most routines)
// v1f4 - reversed declaration order floats 1st boolean last (fixed a lot of "stupid this should be working bugs")
// v1g  - Reworking debounce
// v1h  - Better power saving by pin manipulation during sleep
//      - Miss. bugs with repeated functions (when user presses = again or =-/* then addes another #
// v1h2 - Removed display variable - disp[] write in number decode routine
//      - Separated floats into long and decimal value to help with long decimal rounding errors
// v1i  - fixed decimal place bugs
//      - Added WDT (Watch Dog timer) overflow at 2 seconds reset (reset for "crazy" key mashing errors)
// v1J  - Frustration: Simple operation works - mashing ====== causes - only sometimes an error
// v1k  - Found bug with clearing keys array out of range 
// v1.1 - Release

// -----------------------------------------------------------------------------------------------------------------

#include <avr/sleep.h>
#include <avr/power.h>
#include <avr/wdt.h>
#include <MemoryFree.h>

// -----------------------------------------------------------------------------------------------------------------

#define MaxDigits 5                                    // MaxDigits = Displays - 1
#define DoSerial false
#define useWDT   true

// -----------------------------------------------------------------------------------------------------------------

volatile float firstNumber = 0.0;
volatile float secondNumber = 0.0;

//volatile float tempFLOAT = 0;                       // *** Changed to save memory
//  Reuse  firstNumber inplace

volatile long tempLONG = 0;

long working = 0;

//unsigned long currentMillis;                           // *** Changed to save memory
// Now just using millis()
unsigned long SleepTimer;
// unsigned long SleepLimit = 15000;                    // *** Changed to save memory
#define SleepLimit 15000

volatile int mode = 0;                                        // Must be int, <0 is used

volatile int commandType = 0;                        // Where 1="+", 2="-", 3="/", 4="*", 5="="
volatile int runoncommand = 0;


volatile int tempINT = 0;

byte i = 0;

byte MSDIndex = 0;
byte LSDIndex = 0;
byte LeftIndex = 0;

volatile byte DPDEX = 0;                              // Start at first decimal point

byte currentKey = 0;
// byte result = 0;                                   // *** Changed to save memory                     

byte disp[7] = {
  0, 10, 10, 10, 10, 10};                             // 0 to 9 show, 10 = display off, 11 = "-" symbol, 12 = "E" for error
byte MSD[7]= {
  0, 0, 0, 0, 0, 0};
byte LSD[7]= {
  0, 0, 0, 0, 0, 0};

byte column = 0;
byte rawKeys = 0;
byte KeyValue = 0;
volatile byte ErrorCode = 0;                         // Where 1 = number too big (ie more than 6 digits)
//       2 = WDT overflow
//       3 = Brown-out
//       4 = Switch case jump too high
//       5 = External reset
//       6 = Power On reset

volatile boolean ErrorFlag = false;
boolean negativeFlag = false;
boolean decplace = false;                             // T or F just used to turn DP LED on or off
boolean commandFlag = false;

boolean decimalFlag = false;                          // Decimal place key press
boolean KeyPress = false;
boolean clearkeyval = false;
boolean KeyOK = true;
boolean FirstNumberStart = false;
boolean SecondNumberStart = false;
boolean ClearFlag = false;

boolean secondNegativeFlag = false;
boolean DPsecondNumbFlag = false;

boolean firstNegativeFlag = false;
boolean DPfirstNumbFlag = false;

boolean leadingDig = false;

boolean runoncalc = false;

volatile boolean KeyDown[18];

// -----------------------------------------------------------------------------------------------------------------

void setup() 
{ 
  pinMode(2, INPUT);                                 // CLEAR / ON BUTTON (USED TO WAKE CALCULATOR)
  //  pinMode(3, INPUT_PULLUP);                          // 2nd Function button selector (This pin is currently not connected)

  DDRB = DDRB | 1 << PORTB0;                         // SEGMENT A
  DDRB = DDRB | 1 << PORTB1;                         // SEGMENT B
  DDRB = DDRB | 1 << PORTB2;                         // SEGMENT C 
  DDRB = DDRB | 1 << PORTB3;                         // SEGMENT D
  DDRB = DDRB | 1 << PORTB4;                         // SEGMENT E
  DDRB = DDRB | 1 << PORTB5;                         // SEGMENT F 
  DDRB = DDRB | 1 << PORTB6;                         // SEGMENT G 
  DDRB = DDRB | 1 << PORTB7;                         // SEGMENT DECIMAL PLACE

  pinMode(4, INPUT);                                 // KEYPAD ROW 0 (TOP ROW)
  pinMode(5, INPUT);                                 // KEYPAD ROW 1
  pinMode(6, INPUT);                                 // KEYPAD ROW 2 
  pinMode(7, INPUT);                                 // KEYPAD ROW 3 (BOTTOM ROW)

  digitalWrite(4,LOW);                               // This is just an attemp to reduce power use
  digitalWrite(5,LOW);                               // These pins are pulled down to GND with 10K resistors
  digitalWrite(6,LOW);
  digitalWrite(7,LOW);

  pinMode(A0, OUTPUT);                               // KEYPAD COLUMN 0 (RIGHT MOST) AND DIGIT 0 (RIGHT MOST)
  pinMode(A1, OUTPUT);                               // KEYPAD COLUMN 1 AND DIGIT 1 
  pinMode(A2, OUTPUT);                               // KEYPAD COLUMN 2 AND DIGIT 2
  pinMode(A3, OUTPUT);                               // KEYPAD COLUMN 3 AND DIGIT 3 
  pinMode(A4, OUTPUT);                               // DIGIT 4
  pinMode(A5, OUTPUT);                               // DIGIT 5

#if DoSerial
  Serial.begin(9600);  
#endif

  ClearAllKeys();

  FirstNumberStart = true;

  SleepTimer = millis();

  // Error reporting ------------------------------------------------------------------------------- S
  // Where 1 = number too big (ie more than 6 digits)
  //       2 = WDT overflow
  //       3 = Brown-out
  //       4 = Switch case jump too high
  //       5 = External reset
  //       6 = Power On reset
  //       7 = WDT System reset
  //       9 = Multiple Errors

    tempINT = MCUSR & B00001111;                        // RES,RES,RES,RES,WDRF,BORF,EXTRF,PORF

  if(tempINT > 0)
  { 
    ErrorFlag = true; 
    switch(tempINT)
    {
    case 1:        
      ErrorCode = 6; 
      break;  

    case 2:
      ErrorCode = 5; 
      break;

    case 4:
      ErrorCode = 3; 
      break;

    case 8:
      ErrorCode = 7; 
      break;

    default:
      //      ErrorCode = 9;
      if(bitRead(tempINT,0))
      {
        ErrorCode = 6;                              // Power on
      }
      else
      {
        ErrorCode = 9;                              // Other multiple errors
      }
      break;
    }
  }

  if(bitRead(WDTCSR, WDIF))
  {
    ErrorFlag = true;
    ErrorCode = 7; 
    bitSet(WDTCSR, WDIF);                        // Write a 1 to clear ???
  }

  MCUSR = 0;
  // Error reporting ------------------------------------------------------------------------------- E

#if useWDT
  WDT_Init();
#endif 

}

// -----------------------------------------------------------------------------------------------------------------


void loop() 
{

#if useWDT  
  wdt_reset(); 
#endif

  if((millis() - SleepTimer) > SleepLimit)
  {
    GoToSleep();
  }

  UpdateDisplayCheck();                                    // Check for button press and Update Display LEDs

  if(ErrorFlag)
  {
    mode = 4;
  }

  switch (mode)                                            // Calculator working statemachine
  {
  case 0:                                                  // Building first number
    if(KeyPress)                                           // Any key is pressed
    {
      SleepTimer = millis();
      if(KeyOK == true)                                    // NEW key is pressed
      {
        if(commandFlag == false)                           // Key is not a command
        {
          if(FirstNumberStart && (KeyValue > 0))           // New number being entered, remove leading zero
          {
            BlankDisplay();
            FirstNumberStart = false;
            DPDEX = 0;
            if(decimalFlag)                                // Start with a decimal place
            {
              disp[0] = 0;
            }
          } 

          if(LeftIndex < 6 && !FirstNumberStart && !decimalFlag)                               // Display is not full (max six digits)
          {
            for(i = 6; i >0; i --)
            {
              disp[i] = disp[i-1];
            }
            disp[0] = KeyValue;

            if(DPfirstNumbFlag)
            {
              firstNumber =(float(KeyValue)/pow(10.0,(float(DPDEX)+1.0))) + firstNumber;       // Append key value to number being "built"
            }
            else
            {
              //              firstNumber =(float)KeyValue + (firstNumber*10);                               // Jan 20th
              tempLONG = long(KeyValue) + (tempLONG*10L);
            }

            LeftIndex = LeftIndex +1;

            if(DPfirstNumbFlag)
            {
              DPDEX = DPDEX + 1;
            }

#if DoSerial
            TestPrints();                                                      // Just for testing
#endif  

          }

          if(decimalFlag)
          {
            DPfirstNumbFlag = true;
            // floatFlag = true;
            decimalFlag = false;
          }
        }
        else
        {
          if(FirstNumberStart)
          {
            if(KeyValue ==2)                                                     // Starting to build a negative number
            {
              firstNegativeFlag = true;
              negativeFlag = true;
            }
            // else do nothing - user pressed "silly" command
          }
          else
          {

            //          negativeFlag = false;          
            BlinkFeedback();                                                     // Command key pressed, blink display

            if(KeyValue !=5)                                                     // "=" NOT pressed do nothing
            {
              mode = 1;                                                            // Set command
              firstNumber = float(tempLONG) + firstNumber;                                // Jan 20th - recombine decimal with long
              tempLONG = 0;
              secondNumber = 0.0;                                                  // J21
              SecondNumberStart = true;
              secondNegativeFlag = false;
              LeftIndex = 0;
              commandType = int(KeyValue);
              // Serial.print("Command Pressed !!! :");
              // Serial.println(commandType);
            }
          } 
          commandFlag = false;                                                  

#if DoSerial
          TestPrints();                                                      // Just for testing
#endif 
        }
      }
    }
    break;

  case 1:                                                                         // Building Second number

    //   while(hang)
    //  {
    //    // hang here
    //   }


    if(KeyPress)
    {
      SleepTimer = millis();
      if(KeyOK == true)
      {
        if(commandFlag == false)
        {
          if(SecondNumberStart && (KeyValue > 0))                                 // First digit being entered (but not a zero)
          {
            BlankDisplay();
            SecondNumberStart = false;
            if(!secondNegativeFlag)
            {
              negativeFlag = false;
            }
            DPDEX = 0; 
            if(decimalFlag)                                                        // Start with a decimal place
            {
              disp[0] = 0;
            }
          }  


          if(LeftIndex < 6 && !SecondNumberStart && !decimalFlag)
          {
            for(i = 6; i >0; i --)                                                 // Shift display to left as new numbers added
            {
              disp[i] = disp[i-1];
            }
            disp[0] = KeyValue;


            if(DPsecondNumbFlag)
            {
              secondNumber =(float(KeyValue)/pow(10.0,(float(DPDEX)+1.0))) + secondNumber;                    // Append key value to number being "built"
            }
            else
            {          
              //              secondNumber =(float)KeyValue + (secondNumber*10);
              tempLONG = long(KeyValue) + (tempLONG*10L);
            }


#if DoSerial
            TestPrints();                                                      // Just for testing
#endif 

            LeftIndex = LeftIndex +1;


            if(DPsecondNumbFlag)
            {
              DPDEX = DPDEX + 1;
            }
          }

          if(decimalFlag)
          {
            DPsecondNumbFlag = true;
            decimalFlag = false;
            //floatFlag = true;
          }

        }
        else
        {
          if(SecondNumberStart)
          {
            if(KeyValue ==2)                                                      // Starting to build a negative number
            {
              BlankDisplay();                                                      // Starting with a negative number -> remove firstnumber from display
              disp[0] = 0;
              secondNegativeFlag = true;
              negativeFlag = true;
              commandFlag = false; 
            }                                                                      // User typed a command before any value = do nothing
          }
          else
          {

            BlinkFeedback();                                                      // *** Should test which command is pressed and not just perform "="

            if(KeyValue ==5)
            {
              runoncalc = false;                                                  // = was pressed
            }
            else
            {              
              runoncalc = true;                                                   // Another command was entered, get ready for more work
              runoncommand = int(KeyValue);
            }

            commandFlag = false; 
            mode = 2;
            secondNumber = float(tempLONG) + secondNumber;
            tempLONG = 0;
            LeftIndex = 0;
            BlankDisplay();    
          }  
          commandFlag = false;                                                  // New Jan 16th 2013
        }
      }
    }  

    break;

  case 2:                                                                        // Do Computation   // Where 1="+", 2="-", 3="/", 4="*", 5="=

    if(secondNegativeFlag)
    {
      secondNumber = 0 - secondNumber;
    }

    if(firstNegativeFlag)
    {
      firstNumber = 0 - firstNumber;
    }

#if DoSerial
    Serial.println("Before Calc:");
    Serial.println(firstNumber, 6);
    Serial.println(secondNumber, 6);
#endif

    switch(commandType)
    {
    case 1:                                                        // Addition
      firstNumber = firstNumber+secondNumber;
      break;

    case 2:                                                        // Subraction
      firstNumber = firstNumber-secondNumber;
      break; 

    case 3:                                                        // Division
      firstNumber = firstNumber/secondNumber;
      //floatFlag = true;
      break;      

    case 4:                                                         // Multiplication
      firstNumber = firstNumber * secondNumber;
      break;

    default: 
      firstNumber = 1.0;                                            // Debugging only
      ResetClearAll(); 
      ErrorFlag = true;
      ErrorCode = 4; 
    }

    if((long(firstNumber) > 999999L) || (long(firstNumber) < -99999L))
    {
      //      exponentresult = true;
      ErrorFlag = true;
      ErrorCode = 1;
    }


#if DoSerial
    TestPrints();                                                      // Just for testing
#endif 

    FloatToDigits();                                                 // Always use floats

      //  mode = 3;                                                      // was mode = 0
    //    secondNumber = 0.0;
    //    SecondNumberStart = true;

    if(negativeFlag)
    {
      firstNegativeFlag = true;
    }

    if(runoncalc)                                                   // This happens when user presses +-/* at rather than = after 2nd number
    {
      runoncalc = false;
      commandType = runoncommand;
      secondNumber = 0.0;
      SecondNumberStart = true;                                     
      secondNegativeFlag = false;
      DPsecondNumbFlag = false;
      LeftIndex = 0;
      tempLONG = 0;                                                 // J21
      mode = 1;                                                     // Go back to building second number using command from 1st 2nd number built exi

    }
    else
    {
      mode = 3;                                                     // = was pressed after 2nd number built
      tempLONG = 0;
    }

    break;  

  case 3:                                                            // Using result in next calculation / or just displaying last result and waiting
      if(KeyPress)
    {
      SleepTimer = millis();
      if(KeyOK == true)
      {
        if(commandFlag == true)
        {
          BlinkFeedback();                                           // Command key pressed, blink display
          if(int(KeyValue) == 5)  
          {
            mode = 2;                                                // "=" pressed, redue calc with same last numbers
            tempLONG = 0;
            LeftIndex = 0;
            BlankDisplay();
          }
          else
          {
            commandType = int(KeyValue);
            mode = 1;                                                // Build second number - again, use firstnumber on display
            tempLONG = 0;                                            // J21
            secondNumber = 0.0;
            SecondNumberStart = true;
            secondNegativeFlag = false;
            DPsecondNumbFlag = false;
            // DPDEX = 0;
            LeftIndex = 0;
          }
          commandFlag = false;  
        }
        else                                                          // Re-build a first number
        {
          BlankDisplay();
          FirstNumberStart = false;
          tempLONG = 0;
          mode = 0;
          DPDEX = 0;
          if(decimalFlag)                                             // Start with a decimal place
          {
            disp[0] = 0;
            DPDEX = 1;
            DPfirstNumbFlag = true;
            decimalFlag = false;
          }
          else
          {
            tempLONG = long(KeyValue);                                                        // J21
            //            firstNumber =(float)KeyValue;
            disp[0] = KeyValue;
            LeftIndex = 1;
          }
        }

      }
    } 

    break;

  case 4:

    switch(ErrorCode)
    {
    case 1:                                                      // Number too big or small
/*    
      disp[0] = 21;  
      disp[1] = 22;    
      disp[2] = 17;
      disp[3] = 20;
      disp[4] = 20;
      disp[5] = 19;
*/      
      disp[0] = 11;                                              // rAnGE
      disp[1] = 21;    
      disp[2] = 23;
      disp[3] = 16;
      disp[4] = 12;
      disp[5] = 10; 

      break;

    case 6:                                                      // Power ON Splash Spells out "CAL_1.1"
      disp[0] = 1; 
      disp[1] = 1;    
      disp[2] = 10;
      disp[3] = 15;
      disp[4] = 16;
      disp[5] = 18;    
      DPDEX = 1; 
      break;

    case 3:                                                      // Brown-out = battery low
      disp[0] = 20;
      disp[1] = 15;    
      disp[2] = 10;
      disp[3] = 19;
      disp[4] = 16;
      disp[5] = 17;    
      break;  

    case 5:                                                      // Reset Pin
      disp[0] = 19;
      disp[1] = 11;    
      disp[2] = 5;
      disp[3] = 11;
      disp[4] = 12;
      disp[5] = 10;    
      break; 

    default:                                                      // Show other error number
      disp[0] = ErrorCode;
      disp[1] = 10;    
      disp[2] = 10;
      disp[3] = 12;
      disp[4] = 12;
      disp[5] = 11;
      break; 
    }

    break;

  default:                                                      // Should never get here
    ResetClearAll(); 
    ErrorFlag = true;
    ErrorCode = 4; 
    break;     

  }
}

// -----------------------------------------------------------------------------------------------------------------

void UpdateDisplayCheck()
{

  digitalWrite(column+14, LOW);                              // Turn off current Digit
  PORTB = B11111111;

  column = column +1;
  if(column >MaxDigits)                                      // > 4 or MaxDigits
  {
    column = 0;
  }

  digitalWrite(column+14, HIGH);                             // Power-up Column (both for buttons and LEDs)
  delayMicroseconds(10);                                     // Let keymatrix settle down ?

  delayMicroseconds(600);                                   // Better debouncing ? (Dimmer display ?) 1000 is good


  if(column <4)
  {
    rawKeys = (PIND & B11110000) >>4;                        // Test here for button Press - Read keys port value
    //  rawKeyArray[column] = (PIND & B11110000) >>4; 
    //    oldrawkeys[column] = (PIND & B11110000) >>4;
  }
  else
  {
    rawKeys = 0;
  }
  //  delayMicroseconds(250);                                 // Better debouncing ? (Dimmer display ?)

  if(DPDEX == column)
  {
    decplace = true;
  }
  else
  {
    decplace = false;
  }


  if(disp[column] >9 && !ErrorFlag)                       // Display nothing
  {
    PORTB = B11111111;
    if((column == MaxDigits) && negativeFlag)
    {
      PORTB = B10111111;                                       // Negative symbol
    }
  }
  else
  {    
    //    PORTB = decodeSegments(disp[column], decplace);            // Set 7 Segements on / off 
    //    PORTB = SYMBOLS[disp[column]] ^ decplace <<7; 
    SegmentDecode();      
  }



  //  if(exponentresult)
  //  {

  //  PORTB = B10000110; 
  //   SegmentDecode();
  //   mode = 4;
  //  }

  delayMicroseconds(800);                                      // For a brighter display ? 400 is good

  //  rawKeyArray[column] = (PIND & B11110000) >>4;

  if(rawKeys>0)                                                // Key being pressed
    //  if(rawKeyArray[column]>0 && (rawKeyArray[column] == oldrawkeys[column]))
  {
    decodeKeyPress();
    KeyPress = true;
  }
  else
  {
    KeyPress = false;

    if(column < 4)
    {
      for(i = (column*4); i<((column*4)+4); i++)                 // Clears the KeyDown variable to signal a key release. Only for current column.
      {
        KeyDown[i] = false;
      }
    }
  }

  // delayMicroseconds(250);                                   // For a brighter display ?


  // -----------------------------------------------------------------------------------------------------------------

  //  Test for Clear Key press - this was it's own void before

    clearkeyval = digitalRead(2);

  if(clearkeyval)
  {
    ClearFlag = true;

    SecondNumberStart = true;
    secondNumber = 0.0;    
    secondNegativeFlag = false;
    negativeFlag = false;

    PORTB = B11111111;

#if DoSerial
    TestPrints();                                                      // Just for testing
#endif 

    do                                                                  // Wait for CLEAR key to be released
    {
      delay(100);          
      clearkeyval = digitalRead(2);
    } 
    while (clearkeyval);

    SleepTimer = millis();

    mode = mode - 1;
    if(mode<0 || mode == 2 || ErrorFlag)                          // CLEAR ALL (Interested in mode = 3 but we do a mode --
    {
      ResetClearAll();

      disp[0] = 0;
      for(i = 1; i<6; i++)
      {
        disp[i] = 10;
      }
    }

    secondNumber = 0.0;

    LeftIndex = 0;
    column = 0;
  }
  else
  {
    ClearFlag = false;
  }
}

// -----------------------------------------------------------------------------------------------------------------
void SegmentDecode()
{
  switch (disp[column]) 
  {
  case 0:    
    PORTB =  B11000000;
    break;

  case 1:    
    PORTB =  B11111001;
    break;

  case 2:    
    PORTB =  B10100100;
    break;

  case 3:    
    PORTB =  B10110000;
    break;

  case 4:    
    PORTB =  B10011001;
    break;

  case 5:    
    PORTB =  B10010010;
    break;

  case 6:    
    PORTB =  B10000011;
    break;

  case 7:    
    PORTB =  B11111000;
    break;

  case 8:    
    PORTB =  B10000000;
    break;

  case 9:    
    PORTB =  B10011000;  
    break; 

  case 10:                                                      // Blank space
    PORTB =  B11111111;  
    break;    

  case 11:                                                      // "E"
    PORTB =  B10000110;  
    break; 

  case 12:                                                      // "r"
//    PORTB =  B10101111;  
    PORTB =  B11001110; 
    break; 

  case 13:                                                      // "S"
    PORTB =  B10010010;  
    break; 

  case 14:                                                      // "P"
    PORTB =  B10001100;  
    break;

  case 15:                                                      // "L"
    PORTB =  B11000111;  
    break;

  case 16:                                                      // "A"
    PORTB =  B10001000;  
    break;

  case 17:                                                      // "b"
    PORTB =  B10000011;  
    break;

  case 18:
    PORTB = B11000110;                                            // C  
    break; 

  case 19:
    PORTB = B10000111;                                            // "t"  
    break;   

  case 20:
    PORTB = B10100011;                                            // "o"  
    break;

  case 21:
    //    PORTB = B10010000;                                      // "g"  
    PORTB = B11000010;                                            // 'G"
    break;

  case 22:
        PORTB = B11101111;                                     // "i"  
    break;  

  case 23:
//       PORTB = B10101011;                                     // "n"  
        PORTB = B11001000;                                      // N
    break;  

  default:                                                      // Should never get here  
    PORTB =  B11111111;
    ErrorFlag = true;
    ErrorCode = 4; 
    break;
  }


  if(decplace)
  {
    bitClear(PORTB, 7);
  }
  else
  {
    bitSet(PORTB, 7);
  }

}



/*
int decodeSegments(int numb, boolean DP)
 {
 result = 0;
 
 switch (numb) 
 {
 case 0:    
 result =  B11000000;
 break;
 
 case 1:    
 result =  B11111001;
 break;
 
 case 2:    
 result =  B10100100;
 break;
 
 case 3:    
 result =  B10110000;
 break;
 
 case 4:    
 result =  B10011001;
 break;
 
 case 5:    
 result =  B10010010;
 break;
 
 case 6:    
 result =  B10000011;
 break;
 
 case 7:    
 result =  B11111000;
 break;
 
 case 8:    
 result =  B10000000;
 break;
 
 case 9:    
 result =  B10011000;
 break; 
 
 default:                                                      // Should never get here  
 break;
 }
 
 
 if(DP)
 {
 bitClear(result, 7);
 }
 else
 {
 bitSet(result, 7);
 }
 
 return result;
 }
 */
// -----------------------------------------------------------------------------------------------------------------
/*
void TestForClear()
 {
 clearkeyval = digitalRead(2);
 
 if(clearkeyval)
 {
 ClearFlag = true;
 
 SecondNumberStart = true;
 secondNumber = 0.0;    
 secondNegativeFlag = false;
 negativeFlag = false;
 
 PORTB = B11111111;
 
 #if DoSerial
 TestPrints();                                                      // Just for testing
 #endif 
 
 do                                                                  // Wait for CLEAR key to be released
 {
 delay(100);          
 clearkeyval = digitalRead(2);
 } 
 while (clearkeyval);
 
 mode = mode - 1;
 if(mode<0 || mode == 2 || exponentresult)                          // CLEAR ALL (Interested in mode = 3 but we do a mode --
 {
 ResetClearAll();
 
 disp[0] = 0;
 for(i = 1; i<6; i++)
 {
 disp[i] = 10;
 }
 }
 
 secondNumber = 0.0;
 
 LeftIndex = 0;
 column = 0;
 }
 else
 {
 ClearFlag = false;
 }
 }
 */
// -----------------------------------------------------------------------------------------------------------------

void decodeKeyPress()
{
  tempINT = 0;                                                        // New Dec 20th

  // KeyPad Layout
  //
  // Columns      Rows
  //   3 2 1 0
  //  
  //   7 8 9 /    0
  //   4 5 6 *    1
  //   1 2 3 -    2
  //   . 0 = +    3

  // Value mapping for decoding
  //     12  8  4  0
  //     13  9  5  1
  //     14  10 6  2
  //     15  11 7  3

  // commandType => 1="+", 2="-", 3="/", 4="*", 5="=", 6="Clear"

  for (i=0; i < 4; i++)                                                // This decodes the Rows
  {
    if(bitRead(rawKeys, i))
      //if(bitRead(rawKeyArray[column], i))   
    {
      tempINT = i;
      break;
    }
  } 

  currentKey = (column*4)+tempINT;

  if(KeyDown[currentKey] == false)
  {

    KeyOK = true;
    KeyDown[currentKey] = true;

    switch (currentKey) 
    {
    case 0:    
      KeyValue =  3;
      commandFlag = true;
      break;

    case 1:    
      KeyValue =  4;
      commandFlag = true;
      break;

    case 2:    
      KeyValue =  2;
      commandFlag = true;
      break;

    case 3:    
      KeyValue =  1;
      commandFlag = true;
      break;

    case 4:    
      KeyValue =  9;
      break;

    case 5:    
      KeyValue =  6;
      break;

    case 6:    
      KeyValue =  3;
      break;

    case 7:    
      KeyValue =  5;
      commandFlag = true;
      break;

    case 8:    
      KeyValue =  8;
      break;

    case 9:    
      KeyValue =  5;
      break; 

    case 10:    
      KeyValue =  2;
      break;

    case 11:    
      KeyValue =  11;
      decimalFlag = true;     
      break;  

    case 12:    
      KeyValue =  7;
      break;

    case 13:    
      KeyValue =  4;
      break; 

    case 14:    
      KeyValue =  1;
      break;

    case 15:    
      KeyValue =  0;
      break;  

      //    case 16:                                              // Clear
      //      KeyValue =  6;
      //      commandFlag = true;
      //      break; 
    default:                                                      // Should never get here
      KeyOK = false;
      KeyValue =  0;
      ErrorFlag = true;
      ErrorCode = 4; 
      //  KeyDown[currentKey] = false;
    }
  }
  else
  {
    KeyOK = false;                                                // Not a new key being pressed
  }
}

// -----------------------------------------------------------------------------------------------------------------

void NumberToDigits(long TheNumber)
{
  tempLONG = 0;
  leadingDig = false;

  working = TheNumber;

  tempLONG = working/100000L;   
  tempLONG =  abs(tempLONG); 
  if(tempLONG > 0) {
    leadingDig = true;
    //    disp[5] = byte(tempLONG);
    MSDIndex = 5;
    MSD[6] = byte(tempLONG);
    working = working - (tempLONG*100000L);
  }

  tempLONG = working/10000L;
  tempLONG =  abs(tempLONG); 
  if(tempLONG > 0 || leadingDig) {
    if(!leadingDig)
    {
      MSDIndex = 4;
    }
    leadingDig = true;
    //    disp[4] = byte(tempLONG);
    MSD[5] = byte(tempLONG);
    working = working - (tempLONG*10000L);
  }

  tempLONG = working/1000L;
  tempLONG =  abs(tempLONG); 
  if(tempLONG > 0 || leadingDig) {
    if(!leadingDig)
    {
      MSDIndex = 3;
    }
    leadingDig = true;
    //    disp[3] = byte(tempLONG);
    MSD[4] = byte(tempLONG); 
    working = working - (tempLONG*1000L);
  } 

  tempLONG = working/100L;
  tempLONG =  abs(tempLONG); 
  if(tempLONG > 0|| leadingDig) {
    if(!leadingDig)
    {
      MSDIndex = 2;
    }
    leadingDig = true;
    //    disp[2] = byte(tempLONG);
    MSD[3] = byte(tempLONG);
    working = working - (tempLONG*100L);
  }

  tempLONG = working/10L;
  tempLONG =  abs(tempLONG); 
  if(tempLONG > 0 || leadingDig) {
    if(!leadingDig)
    {
      MSDIndex = 1;
    }
    leadingDig = true;
    //    disp[1] = byte(tempLONG);
    MSD[2] = byte(tempLONG);
    working = working - (tempLONG*10L);
  }

  if(!leadingDig)
  {
    MSDIndex = 0;
  }

  //  disp[0] = byte(working);
  MSD[1] = byte(working);
  tempLONG = 0;
}      

// -----------------------------------------------------------------------------------------------------------------

void DPToDigits(long TheNumber)
{
  tempLONG = 0;

  LSDIndex = 0;

  for(i = 1; i <5; i++)
  {
    LSD[i] = 0;
  }

  working = TheNumber;

  tempLONG = working/1000L;
  tempLONG =  abs(tempLONG); 
  if(tempLONG > 0) {
    LSDIndex = 1;
    LSD[1] = byte(tempLONG);
    working = working - (tempLONG*1000L);
  }

  tempLONG = working/100L;
  tempLONG =  abs(tempLONG); 
  if(tempLONG > 0) {
    LSDIndex = 2;
    LSD[2] = byte(tempLONG); 
    working = working - (tempLONG*100L);
  } 

  tempLONG = working/10L;
  tempLONG =  abs(tempLONG); 
  if(tempLONG > 0) {
    LSDIndex = 3;
    LSD[3] = byte(tempLONG);
    working = working - (tempLONG*10L);
  }

  if(working > 0) {
    LSDIndex = 4;
    LSD[4] = byte(working);
  }

  tempLONG = 0;
}

// -----------------------------------------------------------------------------------------------------------------

void FloatToDigits()
{

#if DoSerial 
  Serial.println("Start Disp update");
  Serial.println(firstNumber, 6);
  Serial.println(secondNumber, 6);
#endif

  if(firstNumber <0)
  {
    negativeFlag = true;
  }
  else
  {
    negativeFlag = false;
  }

  firstNumber = abs(firstNumber);
  //  tempFLOAT = firstNumber;

  tempLONG = 0;

  tempLONG = long(firstNumber);

  NumberToDigits(tempLONG);                                   // Fill MSD[] array with left side of result

  //  tempFLOAT = (firstNumber - long(firstNumber));              // Remove numbers to left of decimal

  //  tempFLOAT = tempFLOAT *10000.0;                             // Should shift decimal value to left

  firstNumber = (firstNumber - long(firstNumber));              // Remove numbers to left of decimal

  firstNumber = firstNumber *10000.0;                             // Should shift decimal value to left  

  //  if((tempFLOAT - long(tempFLOAT)) > 0.5)                     // To Help with Arduino FLOAT rounding errors
  if((firstNumber - long(firstNumber)) > 0.5)                     // To Help with Arduino FLOAT rounding errors
  {
    //    tempLONG = long(tempFLOAT)  + 1L;                         // Remove any extra decimals and round up    
    tempLONG = long(firstNumber)  + 1L;                         // Remove any extra decimals and round up
  }
  else
  {
    //    tempLONG = long(tempFLOAT);                               // Remove any extra decimals    
    tempLONG = long(firstNumber);                               // Remove any extra decimals
  }

  if(tempLONG > 0)                                              // Jan 17th 2013
  {
    DPToDigits(tempLONG);                                       // Fill LSD[] array with right side of result
  }
  else
  {
    LSDIndex = 0;
  }

  for(i = 0; i<6;i++)                                         // Clear the array before reloading it
  {
    disp[i] = 10;
  }

  tempLONG = 0;                                          // Jan 20th
  firstNumber = 0.0;                                     // Jan 20th

  working = 1L;
  // tempLONG = long(MSD[1]);
  for(i = 0; i<=MSDIndex;i++)                            // Start filling display with whole number part of result
  {
    disp[i] = MSD[i+1]; 
    // tempLONG = long(MSD[i+1]) + (tempLONG*(i*10L));
    // tempLONG =  tempLONG + (long(MSD[i+1])*(i*10L)); < - broken - not to the power of ten

    tempLONG =  tempLONG + (long(MSD[i+1])*working);
    working = working * 10L;

#if DoSerial
    Serial.println(disp[i]);
#endif 
  }


  DPDEX = 0;                                             // Place decimal point after integer part of number

#if DoSerial
  Serial.println("LSDIndex");
  Serial.println(LSDIndex);
  Serial.println("MSDIndex");
  Serial.println(MSDIndex);
#endif

  if(LSDIndex>0)                                         // Check if there is a decimal part of the result
  {
    // --    
    tempINT = 6;                                         // Calculate display digits remaining for decimal part
    if(negativeFlag)                                     // Loose one digit to "-" symbol
    {
      tempINT = tempINT -1;
    }  
    tempINT = tempINT - (MSDIndex +1);                   // decimal value can only fill remaining digits

    if(tempINT >0)
    {
      // --    
      // if(tempINT <4 && LSDIndex > tempINT)
      // {
      //   LSDIndex = LSDIndex - (4-tempINT);
      // }

      LSDIndex = min(LSDIndex,tempINT);                  // Fill decimal place with lessor of remaining space or DP digits

      for(i = 1; i<=LSDIndex;i++)
      {
        //        ShiftLeft();

        //       ShiftLeft - S
        for(int i = 5; i >0;i--)
        {
          disp[i] = disp[i-1];
        }
        //        disp[0] = 0;                                   // Not required, we assign "0" value see lines below
        DPDEX = DPDEX +1;
        //       ShiftLeft - E

        disp[0] = LSD[i];

        firstNumber =(float(LSD[i])/pow(10.0,float(i))) + firstNumber;      // Jan 20th
      }
    }
  }
  firstNumber = float(tempLONG) + firstNumber;
  tempLONG = 0;

#if DoSerial  
  Serial.println("END Disp update");
  Serial.println(firstNumber, 6);
  Serial.println(secondNumber, 6);
#endif    
}

// -----------------------------------------------------------------------------------------------------------------
/*
void ShiftLeft()
 {
 for(int i = 5; i >0;i--)
 {
 disp[i] = disp[i-1];
 }
 disp[0] = 0;
 DPDEX = DPDEX +1;
 }
 */
// -----------------------------------------------------------------------------------------------------------------

void BlinkFeedback()
{  
  PORTB = B11111111;
  delay(100);  
}

// -----------------------------------------------------------------------------------------------------------------

void BlankDisplay()
{
  PORTB = B11111111;
  for(i = 0; i <7; i ++)
  {
    disp[i] = 10;
  }
}

// -----------------------------------------------------------------------------------------------------------------

void ClearAllKeys()
{
  for(i =0; i<17; i++)
  {
    KeyDown[i] = false;
  }
}

// -----------------------------------------------------------------------------------------------------------------

void ResetClearAll()
{
  ClearAllKeys();

  PORTB = B11111111;
  for(i = 0; i <7; i ++)
  {
    disp[i] = 10;
  }

  DPfirstNumbFlag = false;
  firstNumber = 0.0;
  DPsecondNumbFlag = false;
  secondNumber = 0.0;
  mode = 0;
  tempINT = 0;
  tempLONG =0;
  //  tempFLOAT = 0.0;
  column = 0;
  rawKeys = 0;
  KeyValue = 0;
  LeftIndex = 0;
  DPDEX = 0;
  // floatFlag = false;
  decplace = false;
  //  exponentresult = false;
  ErrorFlag = false;
  ErrorCode = 0;
  commandFlag = false;
  commandType = 0;
  decimalFlag = false;
  KeyPress = false;
  KeyOK = true;
  FirstNumberStart = true;
  SecondNumberStart = true;
  secondNegativeFlag = false;
  firstNegativeFlag = false;
  ClearFlag = false;
  currentKey = 0;  
  negativeFlag = false;


  MCUSR = 0;                                    // ATmega error flag bits
}

// -----------------------------------------------------------------------------------------------------------------


void TestPrints()
{
  i = 0;                                        // Basic do nothing

#if DoSerial
  Serial.println("");
  Serial.print("First number (4 digits) : ");
  Serial.println(firstNumber, 4);

  Serial.print("First number (6 digits) : ");
  Serial.println(firstNumber, 6);

  Serial.print("TEMP LONG: ");
  Serial.println(tempLONG);

  Serial.print("Error Code :");
  Serial.println(ErrorCode);


  Serial.print("First number LONG: ");
  Serial.println(long(firstNumber));

  Serial.print("Second number: ");
  Serial.println(secondNumber, 4);

  Serial.print("Second number (6 digits) : ");
  Serial.println(secondNumber, 6);

  Serial.print("TEMP LONG: ");
  Serial.println(tempLONG);

  Serial.print("Command: ");
  Serial.print(" - ");
  Serial.println(commandType);

  Serial.print("Mode: ");
  Serial.println(mode);

  Serial.print("Clear: ");
  Serial.println(ClearFlag);

  //  Serial.print("Start: First = ");
  //  Serial.print(FirstNumberStart);      
  //  Serial.print(" : Second = ");
  //  Serial.println(SecondNumberStart); 

  Serial.print("KeyPress: ");
  Serial.println(KeyPress);

  Serial.print("Key OK: ");
  Serial.println(KeyOK);

  Serial.print("Left Index: ");
  Serial.println(LeftIndex);

  Serial.print("DP Index: ");
  Serial.println(DPDEX);

  Serial.print("First Decimal: ");
  Serial.println(DPfirstNumbFlag);

  Serial.print("WDTCSR: ");
  Serial.println(WDTCSR, BIN);

  Serial.print("MCUSR: ");
  Serial.println(MCUSR, BIN);

  Serial.print("freeMemory()=");
  Serial.println(freeMemory());

#endif  
}

// -----------------------------------------------------------------------------------------------------------------

void GoToSleep()
{
#if useWDT  
  wdt_reset();  
  wdt_disable();
#endif  

  PORTC = PORTC | B00111111;                                    // Also set all columns to high (reduce reverse current waste)

  PORTB = B11111111;                                            // Clear all LEDs (Cathod side)

  power_timer1_disable();                                       // 
  power_timer0_disable();                                       // 
  power_timer2_disable();                                       // 

  power_twi_disable();
  power_usart0_disable();
  power_spi_disable();

  ADCSRA &= ~(1<<ADEN);                                         // Turn off analog
  ACSR = (1<<ACD);                                              // Turn off comparator

  attachInterrupt(0, ClearON, RISING);
  set_sleep_mode(SLEEP_MODE_PWR_DOWN);                          //   was   (SLEEP_MODE_EXT_STANDBY);
  cli();

  sleep_enable();

  sei();
  sleep_cpu();
  //    _NOP();
  //    _NOP();
  sleep_disable();

  sei();

#if DoSerial
  power_usart0_enable();
#endif 
  power_timer0_enable();

  detachInterrupt(0);

  PORTC = PORTC & B11000000;                                    // Turn-off oll columns (as to not light any LEDs unintentionaly)
  SleepTimer = millis();

#if useWDT  
  //  wdt_reset();
  //  wdt_enable(WDTO_2S);
  WDT_Init();
#endif  
}

// -----------------------------------------------------------------------------------------------------------------

void ClearON()
{
  ResetClearAll();
}

// -----------------------------------------------------------------------------------------------------------------

#if useWDT

ISR(WDT_vect, ISR_NAKED)                                     // Routine to deal with WDT overflow error 
{   

  cli();                                                      // disable interrupts

  ErrorFlag = true;                                          // Debugging message to user
  ErrorCode = 2; 

  MCUSR = 0;
  WDTCSR &= ~(1<<WDIF);                                      // Clear WDT Int. Flag
  mode = 4;

  sei();
  reti();
}


void WDT_Init()
{
  cli();                                                     //disable interrupts

  wdt_reset();                                               //reset watchdog


  WDTCSR = (1<<WDCE)|(1<<WDE);                               //set up WDT interrupt
  //WDTCSR = (1<<WDIE)|(1<<WDP2)|(1<<WDP1)|(1<<WDP0);                // 2 seconds prescaller ISR only ? 
  WDTCSR = (1<<WDIE)|(1<<WDE)|(1<<WDP2)|(1<<WDP1)|(1<<WDP0);       // 2 seconds prescaller ISR only and Reset
  //WDTCSR = (1<<WDE)|(1<<WDP2)|(1<<WDP1)|(1<<WDP0);                 // 2 seconds ISR only ? - > does not work - resets but hangs??

  MCUSR = 0;
  WDTCSR &= ~(1<<WDE);                                            // Turn off reset


  sei();                                                            //Enable global interrupts
}

#endif 

// -----------------------------------------------------------------------------------------------------------------


